from datetime import datetime
from logging import Logger
from typing import List, Tuple
from threading import Lock
import jsonpickle


class CloudWatchStreamWriter(object):

    def __init__(self, logs_client, group_name: str, stream_name: str, logger: Logger):
        self.logs_client = logs_client
        self.group_name = group_name
        self.stream_name = stream_name
        self.logger = logger
        self._sequence_token = None
        self._lock = Lock()
        self._init_stream()

    def _init_stream(self):
        try:
            # - Backwards compatibility fallback ---------------------------
            # - https://cs18.visualstudio.com/cs18/_workitems/edit/1133

            # check if logging group already exist
            log_groups = self.logs_client.describe_log_groups(logGroupNamePrefix=self.group_name)

            if len(log_groups['logGroups']) == 0:
                # if logging group not exists create it and set retention for 30 days
                self.logs_client.create_log_group(logGroupName=self.group_name)
                self.logs_client.put_retention_policy(logGroupName=self.group_name, retentionInDays=30)

            # - End --------------------------------------------------------

            # check if stream already exist
            log_streams = self.logs_client.describe_log_streams(logGroupName=self.group_name,
                                                                logStreamNamePrefix=self.stream_name)
            if len(log_streams['logStreams']) > 0:
                # if exists get SequenceToken
                log_stream = log_streams['logStreams'][0]
                self._sequence_token = log_stream.get('uploadSequenceToken') or "0"
            else:
                # if not exists create the stream
                self.logs_client.create_log_stream(logGroupName=self.group_name,
                                                   logStreamName=self.stream_name)
                self._sequence_token = "0"
                self.write([(datetime.utcnow(), "Start of log stream")])
                self.logger.info("log stream was created: " + self.stream_name)

        except Exception:
            self.logger.exception("Failed to _init_stream for "+self.stream_name)
            raise

    def _get_stream_details(self, streams):
        streams.update(self.logs_client.describe_log_streams(logGroupName=self.group_name,
                                                             logStreamNamePrefix=self.stream_name))
        # protect in case of failure
        if 'logStreams' not in streams or len(streams['logStreams']) == 0:
            self.logger.warning('log_stream details were not returned by aws ' + str(streams))
            return None

        return streams

    @staticmethod
    def _fix_empty_string(string: str) -> str:
        if len(string) == 0:
            return " "
        return string

    def write(self, log_events: List[Tuple[datetime, str]]):
        """Write log events to AWS CloudWatch Logs service

        Args:
            log_events: list of tuples containing timestamp and log message
        """
        try:
            events = [
                {
                    'timestamp': int(time.timestamp() * 1000),
                    'message': self._fix_empty_string(message)
                } for time, message in log_events]

            with self._lock:
                result = self.logs_client.put_log_events(logGroupName=self.group_name,
                                                         logStreamName=self.stream_name,
                                                         logEvents=events,
                                                         sequenceToken=self._sequence_token)
                if "rejectedLogEventsInfo" in result:
                    raise Exception(result["rejectedLogEventsInfo"])

                self._sequence_token = result['nextSequenceToken']
                if not self._sequence_token:
                    raise Exception('nextSequenceToken is empty. Response: '+jsonpickle.dumps(result))

        except Exception:
            self.logger.exception("Failed to write cloud-log to " + self.stream_name)
            raise



from sidecar.const import Const
from sidecar.kub_api_service import IKubApiService
from sidecar.sandbox_deployment_end_updater import ISandboxDeploymentEndUpdater


class KubSandboxDeploymentEndUpdater(ISandboxDeploymentEndUpdater):
    def __init__(self,
                 kub_api_service: IKubApiService):
        super(KubSandboxDeploymentEndUpdater, self).__init__()
        self.kub_api_service = kub_api_service

    def _set_deployment_end_details(self, deployment_end_status: str):
        annotations = {Const.SANDBOX_DEPLOYMENT_END_STATUS: deployment_end_status}
        self.kub_api_service.update_namespace(annotations)



from logging import Logger

from sidecar.app_instance_identifier import AppInstanceIdentifier
from sidecar.app_instance_identifier_creator import IAppInstanceIdentifierCreator


class AwsAppInstanceIdentifierCreator(IAppInstanceIdentifierCreator):
    def __init__(self, logger: Logger) -> None:
        super().__init__(logger)
        self.logger = logger

    def create(self, app_name: str, ip_address: str, additional_data: {}) -> AppInstanceIdentifier:
        self.logger.info("entered - app_name: '{}', address: {}, additional_data: {}".format(
            app_name,
            ip_address,
            additional_data))

        if 'instance-id' not in additional_data:
            raise Exception('no instance-id found in additional_data send the the request')
        return AppInstanceIdentifier(name=app_name, infra_id=additional_data['instance-id'])

    def create_from_instance_id(self, app_name: str, instance_id: str) -> AppInstanceIdentifier:
        return AppInstanceIdentifier(name=app_name, infra_id=instance_id)



from abc import abstractmethod, ABCMeta
from logging import Logger

from sidecar.app_instance_identifier import AppInstanceIdentifier


class IAppInstanceIdentifierCreator:
    __metaclass__ = ABCMeta

    def __init__(self, logger: Logger) -> None:
        self.logger = logger

    @abstractmethod
    def create(self, app_name: str, ip_address: str, additional_data: {}) -> AppInstanceIdentifier:
        raise NotImplementedError

    @abstractmethod
    def create_from_instance_id(self, app_name: str, instance_id: str) -> AppInstanceIdentifier:
        raise NotImplementedError



from typing import List, Tuple
from datetime import datetime
from abc import ABC, abstractmethod


class LogEntry(object):
    def __init__(self, app: str, instance: str, topic: str, log_events: List[Tuple[datetime, str]]):
        self.app = app
        self.instance = instance
        self.topic = topic
        self.log_events = log_events

    def get_as_string(self) -> str:
        log_events_str = "\n".join(str(time) + " " + message for time, message in self.log_events)
        return "app: " + self.app + "\ninstance: " + self.instance + "\ntopic: " + self.topic + "\n" + log_events_str


class ICloudLogger(ABC):
    @classmethod
    @abstractmethod
    def write(cls, log_entry: LogEntry):
        pass



import os
from sidecar.const import Const
from . import ICloudLogger, LogEntry


class DummyLogger(ICloudLogger):
    def __init__(self, config: dict):
        apps = config.get("apps")
        for app_name in apps:
            app_folder = Const.get_app_folder(app_name)
            if not os.path.exists(app_folder):
                os.mkdir(app_folder)

    def write(self, log_entry: LogEntry):
        app_folder = Const.get_app_folder(log_entry.app)

        filename = "{0}/{1}.{2}.log".format(app_folder, log_entry.instance.replace("docker://", ""), log_entry.topic)
        lines = [line + "\n" for time, line in log_entry.log_events]
        with open(filename, "at") as stream:
            stream.writelines(lines)



import os
import shutil
import tempfile


class FileSystemService:
    def __init__(self):
        pass

    def create_temp_folder(self):
        """
        Create a temporary folder in the os tmp folder.
        :return: The path of the new folder.
        """
        return tempfile.mkdtemp()

    def create_folder(self, folder):
        """
        Create a folder.
        :param str folder: The path of the new folder.
        """
        if not os.path.exists(folder):
            os.makedirs(folder)

    def create_folders(self, path):
        os.makedirs(path)

    def exists(self, path):
        """
        Check if the path exists.
        :param str path: The path to examine
        :rtype bool
        """
        return os.path.exists(path)

    def delete_temp_folder(self, folder):
        """
        Delete a temporary folder from the os tmp folder.
        :param str folder: Folder path.
        """
        shutil.rmtree(folder, ignore_errors=True)

    def create_file(self, path, chmod=None, mode=None):
        """
        Create (or override) a new file.
        :param str path: The path of the new file (example: 'c:\tmp\file.txt}
        """

        if mode is None:
            mode = 'w'

        f = open(path, mode)
        if chmod:
            os.chmod(path, int(chmod))
        return f

    def get_working_dir(self):
        """
        Get the current working directory.
        :rtype: str
        """
        return os.getcwd()

    def get_entries(self, dir):
        """
           Return a list containing the names of the entries in the directory.
           :param str path: The path of of directory to list
           The list is in arbitrary order.  It does not include the special
           entries '.' and '..' even if they are present in the directory.
           """
        return os.listdir(dir)

    def set_working_dir(self, path):
        """
        Set new working directory.
        :type path: str
        """
        os.chdir(path)

    def write_lines_to_file(self, path, lines, chmod):
        with open(path, "w+") as f:
            f.writelines(lines)
        os.chmod(path, chmod)


class WorkingDirectoryScope:
    def __init__(self, file_system: FileSystemService):
        self.file_system = file_system

    def __enter__(self):
        self.folder = self.file_system.create_temp_folder()
        self.prev_working_dir = self.file_system.get_working_dir()
        self.file_system.set_working_dir(self.folder)
        return self.folder

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file_system.set_working_dir(self.prev_working_dir)
        self.file_system.delete_temp_folder(self.folder)



import json
import threading
from logging import Logger
from typing import Dict, List

from sidecar.app_instance_identifier import AppInstanceIdentifier
from sidecar.app_instance_service import IAppInstanceService
from sidecar.app_request_info import AppRequestInfo
from sidecar.const import AppInstanceConfigStatus


class AppConfigurationEndStatus:
    NONE = "none"
    COMPLETED = "completed"
    ERROR = "error"

    def __init__(self, status: str, is_config_ended: bool):
        self.status = status
        self.is_config_ended = is_config_ended

    def is_ended_with_status(self, required_status: str) -> bool:
        return self.is_config_ended and self.status == required_status


class AppRecord:
    def __init__(self, app_name: str, expected_instances_count: int):
        self.app_name = app_name
        self.expected_instances_count = expected_instances_count
        self.instance_config_end_statuses = {}

    def update_instance_config_end_status(self, config_end_status: str, app_instance_identifier: AppInstanceIdentifier):
        # if an end status was already reported from the same infra_id - it will be overridden
        # this is in order to support app container restart on the same pod
        self.instance_config_end_statuses[app_instance_identifier.infra_id] = config_end_status

    def clear_instance_config_end_status(self, app_instance_identifier: AppInstanceIdentifier) -> bool:
        exists = app_instance_identifier.infra_id in self.instance_config_end_statuses
        if exists:
            self.remove_app_instance_identifier(app_instance_identifier=app_instance_identifier)
        return exists

    def clear_all_end_statuses(self):
        self.instance_config_end_statuses.clear()

    def app_configuration_ended(self) -> bool:
        return self.expected_instances_count - len(self.instance_config_end_statuses) == 0

    def get_current_config_end_status(self) -> AppConfigurationEndStatus:
        is_config_ended = self.app_configuration_ended()

        current_config_end_statuses = list(self.instance_config_end_statuses.values())
        if not current_config_end_statuses:
            return AppConfigurationEndStatus(AppConfigurationEndStatus.NONE, is_config_ended)

        if any(status == AppInstanceConfigStatus.ERROR for status in current_config_end_statuses):
            return AppConfigurationEndStatus(AppConfigurationEndStatus.ERROR, is_config_ended)

        if all(status == AppInstanceConfigStatus.COMPLETED for status in current_config_end_statuses):
            return AppConfigurationEndStatus(AppConfigurationEndStatus.COMPLETED, is_config_ended)

        raise Exception("could not calculate app configuration end status from: {}".format(",".join(current_config_end_statuses)))

    def get_ended_app_instance_identifiers(self) -> List[AppInstanceIdentifier]:
        return [AppInstanceIdentifier(self.app_name, infra_id) for infra_id in self.instance_config_end_statuses]

    def remove_app_instance_identifier(self, app_instance_identifier: AppInstanceIdentifier):
        del self.instance_config_end_statuses[app_instance_identifier.infra_id]


class AppsConfigurationEndTracker:
    def __init__(self, logger: Logger, app_requests: List[AppRequestInfo],
                 app_instance_service: IAppInstanceService):
        self._logger = logger
        self._app_instance_service = app_instance_service
        self._lock = threading.RLock()
        self._app_records = self._build_app_records(app_requests, logger)

    def update_app_instance_config_status(self, app_instance_identifier: AppInstanceIdentifier,
                                          app_instance_config_status: str):
        with self._lock:
            self._invalidate_app_records()
            self._update_app_config_status(app_instance_identifier, app_instance_config_status)
            self._log_current_state()

    def all_apps_configuration_ended(self) -> bool:
        with self._lock:
            return all(record.app_configuration_ended() for record in self._app_records.values())

    def all_apps_configuration_ended_with_status(self, required_status: str) -> bool:
        with self._lock:
            return all(record.get_current_config_end_status().is_ended_with_status(required_status)
                       for record in self._app_records.values())

    def get_all_app_configuration_statuses(self) -> Dict[str, AppConfigurationEndStatus]:
        with self._lock:
            return {app_name: app_record.get_current_config_end_status()
                    for app_name, app_record in self._app_records.items()}

    def get_app_configuration_statuses(self, *app_names: str) -> Dict[str, AppConfigurationEndStatus]:
        with self._lock:
            app_statuses = {app_name: self._app_records[app_name].get_current_config_end_status()
                            for app_name in app_names}
            return app_statuses

    @staticmethod
    def _build_app_records(app_requests: List[AppRequestInfo], logger: Logger) -> Dict[str, AppRecord]:
        app_records = {app_request.name: AppRecord(app_name=app_request.name,
                                                   expected_instances_count=app_request.instances_count)
                       for app_request in app_requests}
        return app_records

    def _invalidate_app_records(self):
        # we could have implemented the syncing of the cache by just re-caching all the app instances' end statuses
        # from the cloud provider, but instead we're removing the no longer existing instances and their statuses.
        # this is because sometimes there is a small delay from when the status is set until it is returned on
        # querying the cloud provider, so we might miss some statuses due to timing issues.
        missing_identifiers = self._get_missing_app_instance_identifiers()
        if missing_identifiers:
            for missing_identifier in missing_identifiers:
                self._app_records[missing_identifier.name].remove_app_instance_identifier(missing_identifier)

            self._logger.info("{count} of the configured app instances were purged from the cache: {details}"
                              .format(count=len(missing_identifiers),
                                      details='; '.join(str(app_id) for app_id in missing_identifiers)))

    def _get_missing_app_instance_identifiers(self) -> List[AppInstanceIdentifier]:
        cached_app_instance_identifiers = set(app_instance_id
                                              for app_record in self._app_records.values()
                                              for app_instance_id in app_record.get_ended_app_instance_identifiers())
        if not cached_app_instance_identifiers:
            return []

        actual_app_instance_identifiers = set(self._app_instance_service.get_all_app_instances())
        missing_identifiers = list(cached_app_instance_identifiers - actual_app_instance_identifiers)
        return missing_identifiers

    def _log_current_state(self):
        self._logger.info("the current state of app records:\n{APP_RECORDS}"
                          .format(APP_RECORDS=json.dumps(self._app_records, default=lambda x: x.__dict__, indent=2)))

    def _update_app_config_status(self, app_instance_identifier: AppInstanceIdentifier,
                                  app_instance_config_status: str):
        app_record = self._app_records[app_instance_identifier.name]
        if AppInstanceConfigStatus.is_end_status(app_instance_config_status):
            app_record.update_instance_config_end_status(app_instance_config_status, app_instance_identifier)
        else:
            app_record.clear_instance_config_end_status(app_instance_identifier=app_instance_identifier)
